"""Utilities for parsing a rdata file."""

from ._parser import (
    DEFAULT_ALTREP_MAP as DEFAULT_ALTREP_MAP,
    CharFlags as CharFlags,
    RData as RData,
    RObject as RObject,
    RObjectInfo as RObjectInfo,
    RObjectType as RObjectType,
    parse_data as parse_data,
    parse_file as parse_file,
)
